module mod_communication
	use mpi
	implicit none
contains

	subroutine initialize_array(array_size, array)
		integer, intent(in) :: array_size
		integer :: i
		real, dimension(:), allocatable, intent(out) :: array
		allocate(array(array_size))
		do i=1, array_size	!array with some numbers
			array(i) = 100.0+i**2
		end do
	end subroutine initialize_array

	subroutine calc_partner(rank, numProcesses, partner1, partner2)
		integer, intent(in) :: rank, numProcesses
		integer, intent(out) :: partner1, partner2
		partner1 = rank+1
		partner2 = rank-1
		if (rank == numProcesses-1) then
			partner1 = 0
		else if (rank == 0) then
			partner2 = numProcesses-1
		end if
	end subroutine calc_partner

	subroutine deadlock(numProcesses, rank, array)
		integer, intent(in) :: numProcesses, rank
		real, dimension(:), intent(in) :: array
		integer :: partner1, partner2, array_size, ierr, status(MPI_STATUS_SIZE)
		array_size = size(array)
		call calc_partner(rank, numProcesses, partner1, partner2)
		call MPI_SEND(array, array_size, MPI_REAL, partner1, rank, MPI_COMM_WORLD, ierr)
		call MPI_RECV(array, array_size, MPI_REAL, partner2, partner2, MPI_COMM_WORLD, status, ierr)
	end subroutine deadlock
	
	subroutine no_deadlock(numProcesses, rank, array)
		integer, intent(in) :: numProcesses, rank
		real, dimension(:), intent(in) :: array
		integer :: partner1, partner2, array_size, ierr, status(MPI_STATUS_SIZE)
		array_size = size(array)
		call calc_partner(rank, numProcesses, partner1, partner2)
		if (mod(rank,2) == 0) then
			call MPI_SEND(array, array_size, MPI_REAL, partner1, rank, MPI_COMM_WORLD, ierr)
			call MPI_RECV(array, array_size, MPI_REAL, partner2, partner2, MPI_COMM_WORLD, status, ierr)
		else	!inverse send-receive order for uneven process numbers
			call MPI_RECV(array, array_size, MPI_REAL, partner2, partner2, MPI_COMM_WORLD, status, ierr)
			call MPI_SEND(array, array_size, MPI_REAL, partner1, rank, MPI_COMM_WORLD, ierr)
		end if
	end subroutine no_deadlock

	subroutine non_blocking_communication(numProcesses, rank, array)
		integer, intent(in) :: numProcesses, rank
		real, dimension(:), intent(in) :: array
		integer :: partner1, partner2, array_size, ierr, request, status(MPI_STATUS_SIZE)
		array_size = size(array)
		call calc_partner(rank, numProcesses, partner1, partner2)
		call MPI_ISEND(array, array_size, MPI_REAL, partner1, rank, MPI_COMM_WORLD, request, ierr)
		call MPI_IRECV(array, array_size, MPI_REAL, partner2, partner2, MPI_COMM_WORLD, & 
		&    request, ierr)
		call MPI_WAIT(request, status, ierr)
	end subroutine non_blocking_communication

	subroutine sendrecv(numProcesses, rank, array)
		integer, intent(in) :: numProcesses, rank
		real, dimension(:), intent(in) :: array
		integer :: partner1, partner2, array_size, ierr, status(MPI_STATUS_SIZE)
		array_size = size(array)
		call calc_partner(rank, numProcesses, partner1, partner2)
		call MPI_SENDRECV(array, array_size, MPI_REAL, partner1, rank, array, array_size, MPI_REAL, & 
		&    partner2, partner2, MPI_COMM_WORLD, status, ierr)
	end subroutine sendrecv

end module mod_communication


